﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: CameraReader.cs 537 2012-04-04 15:35:42Z cr333 $
*/
using System.IO;
using System.Linq;
using VideoLib.Vision;

namespace RGBZcamera.Library.Config
{
	/// <summary>Reads camera-related information from an INI file.</summary>
	public class CameraReader
	{
		/// <summary>Gets the intrinsic camera parameters.</summary>
		public IntrinsicParameters IntrParams { get { return ReadIntrinsicParameters(); } }

		/// <summary>Gets the extrinsic camera parameters.</summary>
		public ExtrinsicParameters ExtrParams { get { return ReadExtrinsicParameters(); } }

		/// <summary>Gets a wildcard path describing all frames, e.g. "video-{0:D4}.png".</summary>
		public string Files { get { return Path.Combine(Path.GetDirectoryName(iniFile.Path), iniFile.ReadString(CameraKey, "Filenames")); } }

		/// <summary>Gets the sequence number of the first valid frame.</summary>
		public int FirstFrame { get { return iniFile.ReadInt(CameraKey, "FirstFrame"); } }

		/// <summary>Gets the sequence number of the last valid frame.</summary>
		public int LastFrame { get { return iniFile.ReadInt(CameraKey, "LastFrame"); } }

		/// <summary>Gets the key under which a camera is known in the INI file.</summary>
		public string CameraKey { get; protected set; }

		private IniFile iniFile;
		private string CalibSection { get { return CameraKey + "Calib"; } }

		/// <summary>Initialises a new instance of the <see cref="CameraReader"/> class.</summary>
		/// <param name="filename">The filename of the INI file.</param>
		/// <param name="cameraKey">The key under which the camera is known in the INI file.</param>
		public CameraReader(string filename, string cameraKey)
		{
			iniFile   = new IniFile(filename);
			CameraKey = cameraKey;
		}

		private IntrinsicParameters ReadIntrinsicParameters()
		{
			return new IntrinsicParameters()
			{
				F = iniFile.ReadFloat(CalibSection, "fc"),
				C = iniFile.ReadFloatTuple(CalibSection, "cc"),
				K = iniFile.ReadFloatTuple(CalibSection, "kc")
			};
		}

		private ExtrinsicParameters ReadExtrinsicParameters()
		{
			return new ExtrinsicParameters()
			{
				T = iniFile.ReadFloatTuple(CalibSection, "tc").Select(x => x / 1000).ToArray(), // calibration uses mm, but we use meters
				O = iniFile.ReadFloatTuple(CalibSection, "omc")
			};
		}
	}
}
