﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: RenderQuads.vert 537 2012-04-04 15:35:42Z cr333 $
*/
#version 120
#extension GL_EXT_gpu_shader4 : enable

uniform sampler2D colour;
uniform sampler2D points;
uniform float pointSize;
uniform int input;

varying vec3 vertex;
varying vec3 normal;
varying vec2 direction;

// forward declaration of mesh vertex shader functions (from Mesh.vert)
vec3 getVertex(sampler2D depthmap, vec2 position);
vec3 getVertexNormal(sampler2D depthmap, vec2 position, vec3 coords);


// Reads in tone values for the given channel and pixel position
mat3 readToneImage(sampler2D image, int channel, vec2 position)
{
	float lineWidth = pointSize / 4; // TODO

	// read in tone values
	vec2 texelSize = 1.0 / textureSize2D(image, 0).xy;
	vec4 t1 = texture2D(image, (position + 0.5 * lineWidth * vec2(-1., -1.)) * texelSize);
	vec4 t2 = texture2D(image, (position + 0.5 * lineWidth * vec2( 0., -1.)) * texelSize);
	vec4 t3 = texture2D(image, (position + 0.5 * lineWidth * vec2( 1., -1.)) * texelSize);
	vec4 t4 = texture2D(image, (position + 0.5 * lineWidth * vec2(-1.,  0.)) * texelSize);
	vec4 t5 = texture2D(image, (position + 0.5 * lineWidth * vec2( 0.,  0.)) * texelSize);
	vec4 t6 = texture2D(image, (position + 0.5 * lineWidth * vec2( 1.,  0.)) * texelSize);
	vec4 t7 = texture2D(image, (position + 0.5 * lineWidth * vec2(-1.,  1.)) * texelSize);
	vec4 t8 = texture2D(image, (position + 0.5 * lineWidth * vec2( 0.,  1.)) * texelSize);
	vec4 t9 = texture2D(image, (position + 0.5 * lineWidth * vec2( 1.,  1.)) * texelSize);

	switch(channel)
	{
		case 0: return mat3(t1.x, t2.x, t3.x, t4.x, t5.x, t6.x, t7.x, t8.x, t9.x);
		case 1: return mat3(t1.y, t2.y, t3.y, t4.y, t5.y, t6.y, t7.y, t8.y, t9.y);
		case 2: return mat3(t1.z, t2.z, t3.z, t4.z, t5.z, t6.z, t7.z, t8.z, t9.z);
		case 3: return mat3(t1.w, t2.w, t3.w, t4.w, t5.w, t6.w, t7.w, t8.w, t9.w);
		case 4: return 0.2126 * mat3(t1.x, t2.x, t3.x, t4.x, t5.x, t6.x, t7.x, t8.x, t9.x)
		             + 0.7152 * mat3(t1.y, t2.y, t3.y, t4.y, t5.y, t6.y, t7.y, t8.y, t9.y)
		             + 0.0722 * mat3(t1.z, t2.z, t3.z, t4.z, t5.z, t6.z, t7.z, t8.z, t9.z); // Rec. 709
	}
}

void main(void)
{
	// vertex position and normal, in world coordinates
	vec2 pos = gl_Vertex.xy; // minus one to distribute border evenly on all sides
	vertex = getVertex(points, pos);
	normal = getVertexNormal(points, pos, vertex);

	// default transform of vertex position
	gl_Position = gl_Vertex;//ftransform(); //gl_ModelViewProjectionMatrix * gl_Vertex;
	
	gl_FrontColor = texture2D(colour, gl_Vertex.xy / textureSize2D(colour, 0));
	//gl_FrontColor = vec4(normalize(normal) / 2 + 0.5, 1);
	
	mat3 t = readToneImage(points, input, pos);
	
	// compute coefficents of degree-2 polynomial
	//
	//   f(x,y) = a0 * x^2 + 2 * a1 * x * y + a2 * y^2 + a3 * x + a4 * y + c5
	//
	// using A = H * T (as described in Lee et al. 2007), where H is hardcoded
	// with values determined in Matlab:
	//
	//   [x,y] = meshgrid(-1:1)
	//   y = y'; y = y(:); x = x'; x = x(:);
	//   X = [x .^ 2, 2 * x .* y, y .^ 2, x, y, ones(size(x))]
	//   H = inv(X' * X) * X'
	//   norm(H - [[1 -2 1 1 -2 1 1 -2 1]/6; [1 0 -1 0 0 0 -1 0 1]/8; [1 1 1 -2 -2 -2 1 1 1]/6; ...
	//       [-1 0 1 -1 0 1 -1 0 1]/6; [-1 -1 -1 0 0 0 1 1 1]/6; [-1 2 -1 2 5 2 -1 2 -1]/9])
	//   >> 1.4007e-016
	//float invLineWidth = 1.0 / lineWidth;
	float a0 = ( t[0][0] - 2.0 * t[0][1] + t[0][2] +       t[1][0] - 2.0 * t[1][1] +       t[1][2] + t[2][0] - 2.0 * t[2][1] + t[2][2]) / 6.0;// * invLineWidth * invLineWidth;
	float a1 = ( t[0][0]                 - t[0][2]                                                 - t[2][0]                 + t[2][2]) / 8.0;// * invLineWidth * invLineWidth;
	float a2 = ( t[0][0] +       t[0][1] + t[0][2] - 2.0 * t[1][0] - 2.0 * t[1][1] - 2.0 * t[1][2] + t[2][0] +       t[2][1] + t[2][2]) / 6.0;// * invLineWidth * invLineWidth;
	float a3 = (-t[0][0]                 + t[0][2] -       t[1][0] +                       t[1][2] - t[2][0]                 + t[2][2]) / 6.0;// * invLineWidth;
	float a4 = (-t[0][0] -       t[0][1] - t[0][2]                                 +       t[2][0]           +       t[2][1] + t[2][2]) / 6.0;// * invLineWidth;
	float a5 = (-t[0][0] + 2.0 * t[0][1] - t[0][2] + 2.0 * t[1][0] + 5.0 * t[1][1] + 2.0 * t[1][2] - t[2][0] + 2.0 * t[2][1] - t[2][2]) / 9.0;

	// compute eigenvalues of M = [a0 a1; a1 a2] (see for details:
	// <http://www.math.harvard.edu/archive/21b_fall_04/exhibits/2dmatrices/index.html>)
	// Note: adding small epsilon to work around machine precision (sqrt(negative) = NaN).
	float eigv_term = sqrt(pow((a0 + a2), 2.0) / 4.0 - (a0 * a2 - a1 * a1) + 1e-10);
	float l1 = (a0 + a2) / 2.0 + eigv_term;
	float l2 = (a0 + a2) / 2.0 - eigv_term;
			
	//// reject if the first principal curvature is too low (i.e. locally flat)
	//float c = max(abs(l1), abs(l2));
	//if(c < c_l) return 0.0;
	
	// compute eigenvectors of M, exploiting that M is symmetric.
	vec2 v1, v2;
	if(abs(a1) < 1.e-10) { v1 = vec2(1, 0); v2 = vec2(0, 1); }
	else { v1 = normalize(vec2(l1 - a2, a1)); v2 = normalize(vec2(l2 - a2, a1)); }
	
	if(abs(l1) > abs(l2)) direction = v2;
	else                  direction = v1;
}