﻿/*
This source file is part of the RGBZ Video Camera project published in the paper

  Coherent Spatiotemporal Filtering, Upsampling and Rendering of RGBZ Videos
  Christian Richardt, Carsten Stoll, Neil A. Dodgson, Hans-Peter Seidel, Christian Theobalt
  Computer Graphics Forum (Proceedings of Eurographics), 31(2), May 2012
  <http://www.mpi-inf.mpg.de/resources/rgbz-camera/>

If you use our code, please cite our paper. Thanks.

Copyright (c) 2012 Christian Richardt <christian@richardt.name>


Code licence: GPL v3
----------------------------------------------------------------------------------------------------
This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with this program;
if not, see <http://www.gnu.org/licenses/>.

$Id: Reprojection.frag 537 2012-04-04 15:35:42Z cr333 $
*/
#extension GL_EXT_gpu_shader4 : enable

uniform sampler2D coordinates;

// the vertex position, in viewing coordinates
varying vec3 v3;

varying vec3 n3;
varying float valid;
varying float sobel;
varying vec2 coords;

uniform float sobelThreshold;
uniform float normalThreshold;
uniform int cleanupMode;


vec3 computeFaceNormal(vec2 coords)
{
	// mesh connectivity:
	//
	// TL--TR   top-left/right
	// | \  |
	// |  \ |
	// BL--BR   bottom-left/right
	
	// texture coordinates of triangle vertices
	ivec2 cTL = ivec2(coords);
	ivec2 cTR = ivec2(coords) + ivec2(1, 0);
	ivec2 cBL = ivec2(coords) + ivec2(0, 1);
	ivec2 cBR = ivec2(coords) + ivec2(1, 1);	
	
	// lookup 3D coordinates
	vec2 size = vec2(textureSize2D(coordinates, 0));
	vec3 pTL = texture2D(coordinates, (vec2(cTL) + 0.5) / size).xyz;
	vec3 pBR = texture2D(coordinates, (vec2(cBR) + 0.5) / size).xyz;

	vec3 normal;
	if(fract(coords.x) > fract(coords.y)) // top-right triangle
	{
		vec3 pTR = texture2D(coordinates, (vec2(cTR) + 0.5) / size).xyz;
		return normalize(cross(pBR - pTL, pTR - pTL));
	}
	else // bottom-left triangle
	{
		vec3 pBL = texture2D(coordinates, (vec2(cBL) + 0.5) / size).xyz;
		return normalize(cross(pBL - pTL, pBR - pTL));
	}
}

void main()
{
	//if(n3.z < 0.2) discard;
	
	vec3 normal = computeFaceNormal(coords);
	
	// transform from OpenGL coordinate system (x right, y down, z into the scene)
	// to SR4000 coordinate system (x left, y up, z into the scene), i.e. flip x & z

	// discard invalid fragments (e.g. inside saturated triangles)
	if(valid < 0.999)
	{
		//discard; // previously used, but need xyz for rescaling in filter
		gl_FragColor = vec4(normalize(vec3(-v3.x, v3.y, -v3.z)), 0); // new coordinate format: (x, y, z, distance)
	}
	else if(cleanupMode == 1 && sobel > sobelThreshold) // ignore edge points
	{
		gl_FragColor = vec4(-v3.x, v3.y, -v3.z, 0); // new coordinate format: (x, y, z, distance)
	}
	else if(cleanupMode == 2 && abs(normal.z) < normalThreshold) // reject if face normal points away from viewer
	{
		gl_FragColor = vec4(-v3.x, v3.y, -v3.z, 0); // new coordinate format: (x, y, z, distance)
	}
	else if(cleanupMode == 3 && abs(n3.z) < normalThreshold) // reject if smooth per-pixel normal points away from viewer
	{
		gl_FragColor = vec4(-v3.x, v3.y, -v3.z, 0); // new coordinate format: (x, y, z, distance)
	}
	else
	{
		//// experimental depth cut-off
		//if(length(v3) > 0.8) gl_FragColor = vec4(-v3.x, v3.y, 1, 1); else
		gl_FragColor = vec4(-v3.x, v3.y, -v3.z, length(v3)); // new coordinate format: (x, y, z, distance)
	}
}