﻿// $Id: FlowWarper.cs 494 2010-11-15 20:31:09Z cr333 $
using System.Collections.Generic;
using System.ComponentModel;

namespace VideoLib.OpenGL.OpticalFlow
{
	/// <summary>Defines all warp directions.</summary>
	public enum WarpDirection
	{
		/// <summary>Warps the source image towards the target image.</summary>
		SourceInterpolate,

		/// <summary>Warps the source image away from the target image.</summary>
		SourceExtrapolate,

		/// <summary>Warps the target image towards the source image.</summary>
		TargetInterpolate,

		/// <summary>Warps the target image away from the source image.</summary>
		TargetExtrapolate,
	}

	/// <summary>Warps a texture according to a given flow field.</summary>
	public class FlowWarper : FboNode
	{
		/// <summary>Initialises a new instance of the <see cref="FlowWarper"/> class.</summary>
		public FlowWarper()
		{
			WarpDirection = WarpDirection.SourceInterpolate;
			Magnitude     = 1.0f;

			AddInputPins("image", "flow");
			AddOutputPin("out");
		}

		/// <summary>Gets or sets the direction of the warping process.</summary>
		[Uniform(), Description("The direction of the warping process.")]
		public WarpDirection WarpDirection { get; set; }

		/// <summary>Gets or sets the magnitude of the warp; one means warp from source to target. Default: 1.</summary>
		[Uniform(), DefaultValue(1.0f), Description("The magnitude of the warp; one means warp from source to target.")]
		public float Magnitude { get; set; }

		/// <summary>Initialises the OpenGL state.</summary>
		public override void Init()
		{
			// load, compile and link shaders
			programID = Utils.LinkShaders(
				Utils.CreateShaderFromResource("OpticalFlow.Shaders.FlowWarper.frag")
			);

			// map textures
			textureMap = new List<TextureBinding> {
				new TextureBinding { Name = "image", Unit = 0 },
				new TextureBinding { Name = "flow",  Unit = 1 },
			};

			base.Init();
		}

		/// <summary>Carries out drawing to the framebuffer object using the given input textures.</summary>
		/// <param name="inputs">The input textures.</param>
		/// <returns>The resulting texture.</returns>
		public override TextureFrame Draw(Dictionary<string, TextureFrame> inputs)
		{
			// allocate a texture that is identical to the input
			var resultTexture = TextureFrame.CreateToMatch(inputs["image"]);
			resultTexture.Alloc();

			Draw(inputs, resultTexture.Texture);

			return resultTexture;
		}
	}
}
