/*
This source code is part of the Floating Textures test program, which
is a derivative of the complete Floating Textures program containing
the offline version with prereconstructed geometry and prerecorded
video. 
For more information on Floating Textures please have a look at the article

"Floating Textures" 
by M. Eisemann, B. De Decker, M. Magnor, 
P. Bekaert, E. de Aguiar, N. Ahmed, C. Theobalt, A. Sellent
Computer Graphics Forum (Proc. Eurographics EG'08), vol. 27, no. 2, April 2008 


Copyright (C) 2008 Martin Eisemann

This program is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; either version 3 of the License, or (at your option) any later 
version.

This program is distributed in the hope that it will be useful, but WITHOUT 
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with 
this program; if not, see <http://www.gnu.org/licenses/>.
*/

/**
 * \brief Jacobi.frag
 *
 * \brief Part of the optimization. Calculates the update for DUV
 *
 */ 

// enable draw_buffers //
#extension GL_ARB_draw_buffers : enable


uniform vec2 dxy;

uniform sampler2D texTensor0;
uniform sampler2D texTensor1;
uniform sampler2D texTensor2;
uniform sampler2D texTensor3;

uniform sampler2D texUVField;

uniform sampler2D texPhi3;

// ignore silhouette (CR 2010-08-17)
//uniform sampler2D silhouette;

uniform vec2 alphav2;
uniform vec4 alphav4;
uniform vec4 betav4;

const vec4 epsilonv4 = vec4(0.001, 0.001, 0.001, 0.001);

void invert(in vec2 M11, in vec2 M12, in vec2 M22, inout vec2 iM11, inout vec2 iM12, inout vec2 iM22)
{
  vec2 denom = M11 * M22 - M12 * M12;
  if(denom.x < 0.0001){ // added against division by zero
    denom.x = 0.0001; }
  if(denom.y < 0.0001){
    denom.y = 0.0001; }

  vec2 factor = vec2(1.0, 1.0) / denom;
  
  iM11 = M22  * factor;
  iM12 = -M12 * factor;
  iM22 = M11  * factor;
}

void main()
{ 
  // needed texture coordinates (could be set in application) //
  vec2 tCyx   = gl_TexCoord[0].xy;
  
  // ignore silhouette (CR 2010-08-17)
  //// early termination //
  //if(texture2D(silhouette, tCyx).a == 0.0){
  //  discard;
  //}
  
  vec2 tCyxp1 = tCyx + vec2(dxy.x, 0.0); 
  vec2 tCyp1x = tCyx + vec2(0.0, dxy.y);
  vec2 tCyxm1 = tCyx - vec2(dxy.x, 0.0);
  vec2 tCym1x = tCyx - vec2(0.0, dxy.y);

  // needed values //
  vec4 uvyx   = texture2D(texUVField, tCyx);
  vec4 uvyxp1 = texture2D(texUVField, tCyxp1);
  vec4 uvyxm1 = texture2D(texUVField, tCyxm1);
  vec4 uvyp1x = texture2D(texUVField, tCyp1x);
  vec4 uvym1x = texture2D(texUVField, tCym1x);

  vec4 phi3Fixyx   = texture2D(texPhi3, tCyx); 
  vec4 phi3Fixyxp1 = texture2D(texPhi3, tCyxp1);
  vec4 phi3Fixyp1x = texture2D(texPhi3, tCyp1x);
  vec4 phi3Fixym1x = texture2D(texPhi3, tCym1x);

  vec4 tensor0 = texture2D(texTensor0, tCyx);
  vec4 tensor1 = texture2D(texTensor1, tCyx);
  vec4 tensor2 = texture2D(texTensor2, tCyx);
  vec4 tensor3 = texture2D(texTensor3, tCyx);

  // start computation //
  
  vec2 S11    = tensor0.xz * tensor0.xz;
  vec2 S12_21 = tensor0.xz * tensor0.yw;
  vec2 S13_31 = tensor0.xz * tensor1.xz;
  vec2 S22    = tensor0.yw * tensor0.yw;
  vec2 S23_32 = tensor0.yw * tensor1.xz;
  vec2 S33    = tensor1.xz * tensor1.xz;

  vec2 T11    = (tensor1.yw * tensor1.yw) + (tensor2.xz * tensor2.xz);
  vec2 T12_21 = (tensor1.yw * tensor2.xz) + (tensor2.xz * tensor3.xz);
  vec2 T13_31 = (tensor1.yw * tensor2.yw) + (tensor2.xz * tensor3.yw);
  vec2 T22    = (tensor2.xz * tensor2.xz) + (tensor3.xz * tensor3.xz);
  vec2 T23_32 = (tensor2.xz * tensor2.yw) + (tensor3.xz * tensor3.yw);
  vec2 T33    = (tensor2.yw * tensor2.yw) + (tensor3.yw * tensor3.yw);

  vec4 phi = vec4(S33.x, T33.x, S33.y, T33.y);
  
  
  phi = vec4(0.5, 0.5, 0.5, 0.5) / sqrt(phi + epsilonv4);  

  
  
  // smooth term //
  vec2 smoothTerm = phi3Fixyxp1.xz + vec2(2.0,2.0) * phi3Fixyx.xz + phi3Fixyp1x.xz;

  vec2 M11 = phi.xz * S11    + alphav2 * phi.yw * T11 + smoothTerm;
  vec2 M22 = phi.xz * S22    + alphav2 * phi.yw * T22 + smoothTerm;
  vec2 M12 = phi.xz * S12_21 + alphav2 * T12_21;



  // ru and rv //
  vec4 Sterm = vec4(S13_31.x, S23_32.x, S13_31.y, S23_32.y);
  vec4 Tterm = vec4(T13_31.x, T23_32.x, T13_31.y, T23_32.y);
  vec4 phi2Term = vec4(phi.yw, phi.yw);

  vec4 rurv = -phi * Sterm - alphav4 * phi2Term * Tterm + 

  betav4 * ((phi3Fixyxp1 * (uvyxp1 - uvyx) + phi3Fixyx * (uvyxm1 - uvyx)) + 
	      (phi3Fixyp1x * (uvyp1x - uvyx) + phi3Fixyx * (uvym1x - uvyx)));

  // update du, dv //
  vec2 iM11, iM12, iM22;
  invert(M11, M12, M22, iM11, iM12, iM22);

  vec4 result;
  result.xz = iM11 * rurv.xz + iM12 * rurv.yw;
  result.yw = iM12 * rurv.xz + iM22 * rurv.yw;

  // update UV and DUV for next iteration //
  gl_FragData[0] = uvyx + result; // update uv flow field
  gl_FragData[1] = vec4(0.0, 0.0, 0.0, 0.0); // reset duv flow field
}
