﻿// $Id: ConvertFilter.cs 496 2010-11-16 18:29:58Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>
	/// Simple filter for frame type conversion.
	/// </summary>
	public class ConvertFilter : Node
	{
		/// <summary>Gets or sets the desired output format. Default: None (means same as input).</summary>
		public PixelFormat OutFormat { get; set; }

		/// <summary>Gets or sets if values will be clamped (true) instead of wrapped (false). Default: true (clamped).</summary>
		public bool Clamp { get; set; }

		/// <summary>Gets or sets the scale factor to apply. Default: 1.</summary>
		public float ScaleFactor { get; set; }

		/// <summary>Gets or sets the offset to apply (in 0..1 range). Default: 0.</summary>
		public float Offset { get; set; }

		/// <summary>
		/// Initialises a new instance of the <see cref="ConvertFilter"/> class.
		/// </summary>
		public ConvertFilter()
			: base()
		{
			OutFormat   = PixelFormat.None;
			ScaleFactor = 1.0f;
			Offset      = 0.0f;
			Clamp       = true;

			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>
		/// Initialises a new instance of the <see cref="ConvertFilter"/> class.
		/// </summary>
		/// <param name="targetFormat">The target format.</param>
		public ConvertFilter(PixelFormat targetFormat) : this() { OutFormat = targetFormat; }

		/// <summary>
		/// Processes a video frame.
		/// </summary>
		public override void Process()
		{
			// get the input BitmapFrame
			Frame frame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(frame)) return;
			BitmapFrame inputFrame = CastFrameTo<BitmapFrame>(frame);

			// convert frame format
			BitmapFrame outputFrame = ConvertFormat(inputFrame, OutFormat, ScaleFactor, Offset, Clamp);

			// push output frame out
			PushFrame(outputFrame);
		}

		/// <summary>Converts the input frame to the output format with custom scaling.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outputFormat">The output format.</param>
		/// <param name="scaleFactor">The scale factor.</param>
		/// <param name="offset">The offset applied after scaling.</param>
		/// <param name="clamp">If values are to be clamped to their valid range.</param>
		static public BitmapFrame ConvertFormat(BitmapFrame inputFrame, PixelFormat outputFormat, float scaleFactor, float offset, bool clamp)
		{
			PixelFormat inputFormat = inputFrame.Format;

			// output format 'None' is interpreted as 'same as input'
			if (outputFormat == PixelFormat.None)
				outputFormat = inputFormat;

			// output format just a type? implies same layout as input
			if ((outputFormat & PixelFormat.TypeMask) == outputFormat)
				outputFormat = outputFormat | (inputFormat & ~PixelFormat.TypeMask);

			// handle some special case first (without scaling)
			if (scaleFactor == 1 && offset == 0)
			{
				// no need to convert format?
				if (outputFormat == inputFormat)
					return inputFrame;

				// custom RGB to grey conversion (using luminance)
				if (inputFormat == PixelFormat.BGR24 && outputFormat == PixelFormat.Grey8)
					return ConvertBGR24ToGrey8(inputFrame);
			}

			// conversion of datatypes (same colour components and layout, ignoring alpha)?
			if ((inputFormat & ~PixelFormat.TypeMask & ~PixelFormat.Alpha) == (outputFormat & ~PixelFormat.TypeMask & ~PixelFormat.Alpha))
				return ConvertDatatype(inputFrame, outputFormat, scaleFactor, offset, clamp);

			throw new NotImplementedException(String.Format("Conversion from {0} to {1} not implemented (scale={2}; offset={3}).", inputFormat, outputFormat, scaleFactor, offset));
		}

		/// <summary>Converts the input frame to the output format.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outputFormat">The output format.</param>
		static public BitmapFrame ConvertFormat(BitmapFrame inputFrame, PixelFormat outputFormat)
		{
			return ConvertFormat(inputFrame, outputFormat, 1, 0, true);
		}

		/// <summary>Converts a frame from one to another datatype (wrap around if under-/overflow).</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <param name="outputFormat">The output format.</param>
		/// <param name="scaleFactor">The scale factor.</param>
		/// <param name="offset">The offset applied after scaling.</param>
		/// <param name="clamp">If values are to be clamped to their valid range.</param>
		private static BitmapFrame ConvertDatatype(BitmapFrame inputFrame, PixelFormat outputFormat, float scaleFactor, float offset, bool clamp)
		{
			BitmapFrame outputFrame = new BitmapFrame(inputFrame.Width, inputFrame.Height, outputFormat) { SequenceNumber = inputFrame.SequenceNumber };

			PixelFormat inputType  = inputFrame.Format & PixelFormat.TypeMask;
			PixelFormat outputType = outputFormat & PixelFormat.TypeMask;
			int componentsPerRow = inputFrame.Width * inputFrame.Format.ComponentsPerPixel();

			// compute per-colour-component scale factor (just once)
			scaleFactor = (float)(scaleFactor / inputType.MaxValue() * outputType.MaxValue());
			offset      = (float)(offset * outputType.MaxValue());

			// min/max values for clamping
			float minValue = (float)outputType.MinValue();
			float maxValue = (float)outputType.MaxValue();

			unsafe
			{
				// pin the input & output pointers
				fixed (byte* pIn = inputFrame.Data, pOut = outputFrame.Data)
				{
					for (int y = 0; y < inputFrame.Height; y++)
					{
						// compute row pointers for both images (in bytes)
						byte* rIB = pIn + y * inputFrame.Stride;
						byte* rOB = pOut + y * outputFrame.Stride;

						// convert pointers to different datatypes
						ushort* rIS = (ushort*)rIB;
						uint*   rII = (uint  *)rIB;
						float*  rIF = (float *)rIB;
						ushort* rOS = (ushort*)rOB;
						uint*   rOI = (uint  *)rOB;
						float*  rOF = (float *)rOB;

						if (clamp)
						{
							if      (inputType == PixelFormat.Byte   && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(ConvertFilter.clamp(*rIB++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(ConvertFilter.clamp(*rIS++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(ConvertFilter.clamp(*rII++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(ConvertFilter.clamp(*rIF++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Byte   && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(ConvertFilter.clamp(*rIB++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(ConvertFilter.clamp(*rIS++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(ConvertFilter.clamp(*rII++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(ConvertFilter.clamp(*rIF++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Byte   && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(ConvertFilter.clamp(*rIB++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(ConvertFilter.clamp(*rIS++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(ConvertFilter.clamp(*rII++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(ConvertFilter.clamp(*rIF++ * scaleFactor + offset + 0.5f, minValue, maxValue)); } // round
							else if (inputType == PixelFormat.Byte   && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = ConvertFilter.clamp(*rIB++ * scaleFactor + offset, minValue, maxValue); }
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = ConvertFilter.clamp(*rIS++ * scaleFactor + offset, minValue, maxValue); }
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = ConvertFilter.clamp(*rII++ * scaleFactor + offset, minValue, maxValue); }
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = ConvertFilter.clamp(*rIF++ * scaleFactor + offset, minValue, maxValue); }
							else throw new NotImplementedException(String.Format("Conversion from \"{0}\" to \"{1}\" not implemented.", inputFrame.Format, outputFormat));
						}
						else // wrap values, i.e. underflow or overflow
						{
							if      (inputType == PixelFormat.Byte   && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(*rIB++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(*rIS++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(*rII++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Byte  ) { for (int x = 0; x < componentsPerRow; x++) *rOB++ = (byte)(*rIF++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Byte   && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(*rIB++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(*rIS++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(*rII++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Short ) { for (int x = 0; x < componentsPerRow; x++) *rOS++ = (ushort)(*rIF++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Byte   && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(*rIB++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(*rIS++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(*rII++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Int   ) { for (int x = 0; x < componentsPerRow; x++) *rOI++ = (uint)(*rIF++ * scaleFactor + offset + 0.5f); } // round
							else if (inputType == PixelFormat.Byte   && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (*rIB++ * scaleFactor + offset); }
							else if (inputType == PixelFormat.Short  && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (*rIS++ * scaleFactor + offset); }
							else if (inputType == PixelFormat.Int    && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (*rII++ * scaleFactor + offset); }
							else if (inputType == PixelFormat.Single && outputType == PixelFormat.Single) { for (int x = 0; x < componentsPerRow; x++) *rOF++ = (*rIF++ * scaleFactor + offset); }
							else throw new NotImplementedException(String.Format("Conversion from \"{0}\" to \"{1}\" not implemented.", inputFrame.Format, outputFormat));
						}
					}
				}
			}

			return outputFrame;
		}

		private static float clamp(float value, float min, float max)
		{
			if(value < min) return min;
			if(value > max) return max;
			return value;
		}

		/// <summary>Converts an BGR24 frame to 8-bit greyscale.</summary>
		/// <param name="inputFrame">The input frame.</param>
		/// <returns>The output frame.</returns>
		private static BitmapFrame ConvertBGR24ToGrey8(BitmapFrame inputFrame)
		{
			// first allocate the output frame
			BitmapFrame outputFrame = new BitmapFrame(inputFrame.Width, inputFrame.Height, PixelFormat.Grey8) { SequenceNumber = inputFrame.SequenceNumber };

			unsafe
			{
				// pin the input & output pointers
				fixed (byte* pIn = inputFrame.Data, pOut = outputFrame.Data)
				{
					for (int y = 0; y < inputFrame.Height; y++)
					{
						// compute row pointers for both images
						byte* rIn = pIn + y * inputFrame.Stride;
						byte* rOut = pOut + y * outputFrame.Stride;

						// copy it across
						for (int x = 0; x < inputFrame.Width; x++)
						{
							// NOTE: BGR memory layout
							//*rOut = (byte)((rIn[0] + rIn[1] + rIn[2]) / 3.0f + 0.5f); // simple averaging
							//*rOut = (byte)(0.1144845f * rIn[0] + 0.5865990f * rIn[1] + 0.2989164f * rIn[2] + 0.5f); // NTSC
							*rOut = (byte)(0.0721750f * rIn[0] + 0.7151522f * rIn[1] + 0.2126729f * rIn[2] + 0.5f); // sRGB
							rIn += 3;
							rOut++;
						}
					}
				}
			}

			return outputFrame;
		}

		/// <summary>Returns a <see cref="System.String"/> that represents this instance.</summary>
		/// <returns>A <see cref="System.String"/> that represents this instance.</returns>
		public override string ToString()
		{
			var datatype = OutFormat.GetDataType();

			// format scale and offset
			var scale  = (ScaleFactor == 1 ? "" : ScaleFactor.ToString());
			var offset = "";
			if(Offset < 0) offset = "-" + Math.Abs(Offset).ToString();
			if(Offset >= 0) offset = "+" + Math.Abs(Offset).ToString();

			if(datatype == PixelFormat.None)
				return String.Format("{0}({1}x{2}, {3})", Name, scale, offset, (Clamp ? "clamp" : "wrap"));

			return String.Format("{0}({1}, {2}x{3}, {4})", Name, OutFormat.GetDataType(), scale, offset, (Clamp ? "clamp" : "wrap"));
		}
	}
}
