﻿// $Id: CropFilter.cs 350 2010-08-24 21:27:00Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using VideoLib.Frames;
using System.Diagnostics;

namespace VideoLib.Filters
{
	/// <summary>Crops a video frame.</summary>
	public class CropFilter : Node
	{
		// There are 2 ways how the cropping of a frame can be described:
		//
		//   1. In absolute terms, fixing the width and/or height of the cropping rectangle.
		//      This causes the cropped result to always have the same size -- except if either
		//      the width or the height are zero, in which case the maximum possible width or
		//      height will be used.  Use Size for this.
		//
		//   2. In relative terms, relative to the bottom-right corner.
		//      This makes the cropping rectangle dependent on the size of the input frame,
		//      allowing things such as always cropping off a strip of 16 pixels around a frame.
		//      Use BottomRight for this.
		//
		// These 2 ways are exclusive, i.e. setting one clears the other.
		//
		// In addition, the position of the top-left corner can be defined using TopLeft.
		// For both TopLeft and BottomRight, positive values will crop pixels, whereas negative
		// values will add black padding on the respective side(s).

		/// <summary>Cropping relative to the top-left corner of input frame.</summary>
		/// <remarks>Positive values cause cropping, negative values cause black padding.</remarks>
		public Point TopLeft { get { return topLeft; } set { topLeft = value; } }
		private Point topLeft = Point.Empty;

		/// <summary>Cropping relative to the bottom-right corner of input frame. Set clears Size.</summary>
		/// <remarks>Positive values cause cropping, negative values cause black padding.</remarks>
		public Point BottomRight { get { return bottomRight; } set { bottomRight = value; size = Size.Empty; } }
		private Point bottomRight = Point.Empty;

		/// <summary>Absolute size of cropping region. Set clears BottomRight.</summary>
		/// <remarks>Zeros in either width or height are interpreted as using
		/// the maximum possible resolution for that direction.</remarks>
		public Size Size { get { return size; } set { size = value; bottomRight = Point.Empty; } }
		private Size size = Size.Empty;
		
		// Convenience getters/setters ------------------------------------------------------------

		/// <summary>Left cropping.</summary>
		/// <remarks>Positive values cause cropping, negative values cause black padding.</remarks>
		public int Left
		{
			get { return topLeft.X; }
			set { TopLeft = new Point(value, topLeft.Y); }
		}

		/// <summary>Top cropping.</summary>
		/// <remarks>Positive values cause cropping, negative values cause black padding.</remarks>
		public int Top
		{
			get { return topLeft.Y; }
			set { TopLeft = new Point(topLeft.X, value); }
		}

		/// <summary>Right cropping.</summary>
		/// <remarks>Set clears Size. Positive values cause cropping, negative values cause black padding.</remarks>
		public int Right
		{
			get { return bottomRight.X; }
			set { BottomRight = new Point(value, bottomRight.Y); }
		}

		/// <summary>Bottom cropping.</summary>
		/// <remarks>Set clears Size. Positive values cause cropping, negative values cause black padding.</remarks>
		public int Bottom
		{
			get { return bottomRight.Y; }
			set { BottomRight = new Point(bottomRight.X, value); }
		}

		/// <summary>Absolute width of the cropping.</summary>
		/// <remarks>Set clears BottomRight. Zero means maximum width.</remarks>
		public int Width
		{
			get { return size.Width; }
			set { Size = new Size(value, size.Height); }
		}

		/// <summary>Absolute height of the cropping.</summary>
		/// <remarks>Set clears BottomRight. Zero means maximum height.</remarks>
		public int Height
		{
			get { return size.Height; }
			set { Size = new Size(size.Width, value); }
		}

		/// <summary>Sets absolute clipping parameters in one go.</summary>
		public Rectangle AbsRect { set { TopLeft = value.Location; Size = value.Size; } }

		/// <summary>Sets relative clipping parameters in one go.</summary>
		public Rectangle RelRect { set { TopLeft = value.Location; BottomRight = (Point)value.Size; } }

		/// <summary>Constructs a new cropping filter.</summary>
		public CropFilter()
		{
			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>Processes a video frame.</summary>
		public override void Process()
		{
			// get input frame
			Frame inputFrame = GetInputFrame("in");
			if (PassOnEndOfStreamFrame(inputFrame)) return;
			BitmapFrame iFrame = CastFrameTo<BitmapFrame>(inputFrame);

			// compute absolute crop rectangle and apply the crop
			Rectangle crop = CalculateCropRectangle(iFrame.Width, iFrame.Height);
			BitmapFrame oFrame = Apply(iFrame, crop);
			PushFrame(oFrame);
		}

		/// <summary>Calculates the final crop rectangle given some frame size.</summary>
		/// <param name="frameWidth">Frame width.</param>
		/// <param name="frameHeight">Frame height.</param>
		/// <returns>Crop rectangle in terms of pixel coordinates in the input frame.</returns>
		private Rectangle CalculateCropRectangle(int frameWidth, int frameHeight)
		{
			Rectangle rect = new Rectangle(0, 0, frameWidth, frameHeight);
			if (Size != Size.Empty)
			{
				// compute crop rectangle using given absolute size
				// if width or height are zero, then use maximum width/height
				int width  = Size.Width  > 0 ? Size.Width  : frameWidth  - Left;
				int height = Size.Height > 0 ? Size.Height : frameHeight - Top;
				rect = new Rectangle(Left, Top, width, height);
			}
			else if (BottomRight != Point.Empty)
			{
				// compute crop rectangle using given relative offset from bottom-right point
				Size cropSize = new Size(frameWidth, frameHeight) - (Size)TopLeft - (Size)BottomRight;
				rect = new Rectangle(TopLeft, cropSize);
			}
			else
			{
				// compute crop rectangle using only the top-left point
				rect = new Rectangle(TopLeft, new Size(frameWidth, frameHeight) - (Size)TopLeft);
			}

			// sanity check
			Debug.Assert(rect.Left < rect.Right, "Crop has positive width.");
			Debug.Assert(rect.Top < rect.Bottom, "Crop has positive height.");

			return rect;
		}

		/// <summary>Crops the given frame according to the specified cropping region.</summary>
		/// <remarks>Supports negative crop which just results in black padding.</remarks>
		/// <param name="source">The source frame.</param>
		/// <param name="crop">The crop region relative to the top-left pixel.</param>
		/// <returns>A bitmap frame of the size specified by <c>crop.Size</c>.</returns>
		public static BitmapFrame Apply(BitmapFrame source, Rectangle crop)
		{
			// create output frame
			BitmapFrame target = new BitmapFrame(crop.Width, crop.Height, source.Format) { SequenceNumber = source.SequenceNumber };

			// copy in the cropped contents
			int bpp = source.Format.BytesPerPixel();
			for (int yOut = 0; yOut < target.Height; yOut++)
			{
				int yIn = crop.Top + yOut;
				if (yIn >= 0 && yIn < source.Height) // handles negative crop at top & bottom
				{
					// handle negative cropping on the left
					int xIn  = crop.Left >= 0 ? crop.Left : 0;
					int xOut = crop.Left >= 0 ? 0 : -crop.Left;

					Array.Copy(
						source.Data, yIn  * source.Stride + bpp * xIn,
						target.Data, yOut * target.Stride + bpp * xOut,
						bpp * Math.Min(source.Width - xIn, crop.Width)); // handles negative crop on the right
				}
			}

			return target;
		}
	}
}
