﻿// $Id: DelayFilter.cs 387 2010-09-17 16:17:17Z cr333 $
using System.Collections.Generic;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>A node that delays input frames by a given amount.</summary>
	public class DelayFilter : Node
	{
		/// <summary>Gets or sets the delay, in number of frames. Default: 1.</summary>
		public int Delay { get; set; }

		/// <summary>Gets or sets whether the first frame is duplicated while the queue fills up. Default: off.</summary>
		public bool DuplicateFirst { get; set; }

		/// <summary>Gets or sets whether output frames are renumbered to match the input frames. Default: false.</summary>
		public bool Renumber { get; set; }

		/// <summary>Contains the frames that are being delayed.</summary>
		private Queue<Frame> queuedFrames = new Queue<Frame>();

		/// <summary>Initialises a new instance of the <see cref="DelayFilter"/> class.</summary>
		public DelayFilter() : this(1)
		{
		}
		
		/// <summary>Initialises a new instance of the <see cref="DelayFilter"/> class.</summary>
		/// <param name="delay">The delay, in number of frames.</param>
		public DelayFilter(int delay) : base()
		{
			Delay          = delay;
			DuplicateFirst = false;
			Renumber       = false;

			AddInputPin("in");
			AddOutputPin("out");
		}

		/// <summary>Processes some frames, to produce new frames.</summary>
		public override void Process()
		{
			// append input frame to queue of frames
			Frame inputFrame = GetInputFrame("in");
			queuedFrames.Enqueue(inputFrame);

			// determine which frame to output
			Frame outputFrame = null;
			bool needToFree = false;
			if (queuedFrames.Count > Delay)
			{
				// take first from queue
				outputFrame = queuedFrames.Dequeue();
				needToFree  = true;
			}
			else if (DuplicateFirst)
			{
				// duplicate first frame
				outputFrame = queuedFrames.Peek();
				needToFree  = false;
			}

			if (outputFrame != null)
			{
				if (PassOnEndOfStreamFrame(outputFrame))
					return;

				if (Renumber)
					outputFrame.SequenceNumber = inputFrame.SequenceNumber;

				PushFrame(outputFrame);

				if (needToFree)
				{
					// decrement reference count, if applicable
					IRefCountedFrame refCounted = outputFrame as IRefCountedFrame;
					if (refCounted != null) refCounted.RefDec();
				}
			}
		}
	}
}
