﻿// $Id: SyncFilter.cs 436 2010-10-15 16:26:38Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VideoLib.Frames;

namespace VideoLib.Filters
{
	/// <summary>Implements a filter that synchronises a number of input streams, to output their frames simultanously.</summary>
	public class SyncFilter : Node
	{
		/// <summary>Number of streams to synchronise.</summary>
		public int N
		{
			get { return _n; }
			protected set {
				lock (this) // don't modify this node while it's running
				{
					// TODO: delete IO pins if N decreases
					// TODO: add IO pins if N decreases
					_n = value;
					Name = String.Format("SyncFilter({0})", _n);
				}
			}
		}
		private int _n = 0;

		/// <summary>Constructs a new synchronisation filter.</summary>
		/// <param name="numStreams">Number of streams to synchronise.</param>
		public SyncFilter(int numStreams) : base()
		{
			N = numStreams;
			for (int i = 0; i < N; i++)
			{
				AddInputPin("in" + i);
				AddOutputPin("out" + i);
			}
		}

		/// <summary>Processes a set of frames.</summary>
		public override void Process()
		{
			// get input frames
			var frames = new Dictionary<int, Frame>();
			foreach(var i in Enumerable.Range(0, N))
				frames.Add(i, GetInputFrame("in" + i));

			// pass on any end-of-stream frames
			// TODO: be more flexible here, e.g. ignore streams in UpdateStatus that ended previously
			if(frames.Select(x => PassOnEndOfStreamFrame(x.Value)).Any(x => x))
				return;

			// output the frames
			foreach (var frame in frames)
			{
				PushFrame("out" + frame.Key, frame.Value);

				// decrement reference count, if applicable
				IRefCountedFrame refCounted = frame.Value as IRefCountedFrame;
				if (refCounted != null) refCounted.RefDec();
			}
		}
	}
}
