﻿// $Id: ParameterInfo.cs 156 2010-02-18 11:02:58Z cr333 $
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;

namespace VideoLib.Parameters
{
    /// <summary>
    /// Represents a set of details about a parameter (a CLR property marked with [<see cref="ParameterAttribute"/>]),
    /// obtained using reflection on a certain type.
    /// </summary>
    [Serializable]
    public struct ParameterInfo : IEquatable<ParameterInfo>
    {
        private readonly PropertyInfo propInfo;
        private readonly MethodInfo propertyTypeComparer;
        private readonly string name;
        private readonly string description;
        private readonly string groupId;
        private readonly object minVal, maxVal, defaultVal;

        /// <summary>
        /// Creates and initializes an instance of ParameterInfo, with 'null' settings for range and default value.
        /// </summary>
        /// <param name="pinfo">The property info for the parameter's underlying CLR property.</param>
        /// <param name="theName">The friendly name of the parameter.</param>
        /// <param name="theDescription">A description/explanation of the parameter.</param>
        /// <param name="groupName">A 'group id' for identifying groups of similar parameters.</param>
        public ParameterInfo(PropertyInfo pinfo, string theName, string theDescription, string groupName)
        {
            propInfo = pinfo;
            name = theName;
            description = theDescription;
            groupId = groupName;
            minVal = maxVal = defaultVal = null;

            propertyTypeComparer = null;
            try
            {
                InterfaceMapping map = propInfo.PropertyType.GetInterfaceMap(
                    typeof(IComparable<>).MakeGenericType(propInfo.PropertyType));
                
                propertyTypeComparer = map.InterfaceMethods[0];
            }
            catch (ArgumentException) { }
        }

        /// <summary>
        /// Creates and initializes an instance of ParameterInfo.
        /// </summary>
        /// <param name="pinfo">The property info for the parameter's underlying CLR property.</param>
        /// <param name="theName">The friendly name of the parameter.</param>
        /// <param name="theDescription">A description/explanation of the parameter.</param>
        /// <param name="groupName">A 'group id' for identifying groups of similar parameters.</param>
        /// <param name="min">The minimum allowable value for the parameter.</param>
        /// <param name="max">The maximum allowable value for the parameter.</param>
        /// <param name="theDefault">The default value of the parameter.</param>
        public ParameterInfo(PropertyInfo pinfo, string theName, string theDescription, string groupName, object min, object max, object theDefault)
        {
            propInfo = pinfo;
            name = theName;
            description = theDescription;
            groupId = groupName;
            minVal = min;
            maxVal = max;
            defaultVal = theDefault;

            propertyTypeComparer = null;
            try
            {
                InterfaceMapping map = propInfo.PropertyType.GetInterfaceMap(
                    typeof(IComparable<>).MakeGenericType(propInfo.PropertyType));

                propertyTypeComparer = map.InterfaceMethods[0];
            }
            catch (ArgumentException) { }
        }

        #region Equality comparison and GetHashCode
        /// <summary>
        /// Checks for equality of the underlying property info object (type and property name)
        /// </summary>
        /// <param name="other">Another ParameterInfo to compare to</param>
        /// <returns><c>true</c> if other and this instance wrap the same property, otherwise <c>false</c></returns>
        public bool Equals(ParameterInfo other)
        {
            return propInfo.DeclaringType.Equals(other.propInfo.DeclaringType) &&
                propInfo.Name.Equals(other.propInfo.Name);
        }
        /// <summary>
        /// Checks for equality of the underlying PropertyInfo object (type and property name)
        /// </summary>
        /// <param name="obj">Another object to compare to</param>
        /// <returns><c>true</c> if obj and this instance are the same type and wrap the same property, otherwise <c>false</c></returns>
        public override bool Equals(object obj)
        {
            if (!(obj is ParameterInfo))
                return false;

            ParameterInfo other = (ParameterInfo)obj;
            return propInfo.DeclaringType.Equals(other.propInfo.DeclaringType) &&
                propInfo.Name.Equals(other.propInfo.Name);
        }
        /// <summary>
        /// Return a hash code consistent with <see cref="Equals(object)"/>
        /// </summary>
        /// <returns>A 32-bit signed integer that is the hash code for this instance</returns>
        public override int GetHashCode()
        {
            return propInfo.DeclaringType.GetHashCode() ^ propInfo.Name.GetHashCode();
        }
        /// <summary>
        /// Checks for equality of the two ParameterInfo instances passed (using underlying property declared type and name)
        /// </summary>
        /// <param name="p0">First instance</param>
        /// <param name="p1">Second instance</param>
        /// <returns><c>true</c> if other and this instance wrap the same property, otherwise <c>false</c></returns>
        public static bool operator ==(ParameterInfo p0, ParameterInfo p1)
        {
            return p0.Equals(p1);
        }
        /// <summary>
        /// Checks for inequality of the two ParameterInfo instances passed (using underlying property declared type and name)
        /// </summary>
        /// <param name="p0">First instance</param>
        /// <param name="p1">Second instance</param>
        /// <returns><c>true</c> if other and this instance do not wrap the same property, otherwise <c>false</c></returns>
        public static bool operator !=(ParameterInfo p0, ParameterInfo p1)
        {
            return !p0.Equals(p1);
        }
        #endregion

        /// <summary>
        /// Gets the type of the parameter's underlying property .
        /// </summary>
        public Type ParamType
        {
            get { return propInfo.PropertyType; }
        }

        /// <summary>
        /// Gets the type containing the parameter.
        /// </summary>
        public Type TargetType
        {
            get { return propInfo.DeclaringType; }
        }

        /// <summary>
        /// Gets the friendly name of the parameter.
        /// </summary>
        public string Name
        {
            get { return name; }
        }

        /// <summary>
        /// Gets a description/explanation of the parameter (if given, otherwise null).
        /// </summary>
        public string Description
        {
            get { return description; }
        }

        /// <summary>
        /// Gets a group identified for the parameter (if given, otherwise null).
        /// </summary>
        public string GroupName
        {
            get { return groupId; }
        }

        /// <summary>
        /// Gets the minimum allowable value of the parameter (if given, otherwise null).
        /// </summary>
        public object Minimum
        {
            get { return (minVal == null) ? null : Convert.ChangeType(minVal, propInfo.PropertyType); }
        }

        /// <summary>
        /// Gets the maximum allowable value of the parameter (if given, otherwise null).
        /// </summary>
        public object Maximum
        {
            get { return (maxVal == null) ? null : Convert.ChangeType(maxVal, propInfo.PropertyType); }
        }

        /// <summary>
        /// Gets the default value of the parameter (if given, otherwise null).
        /// </summary>
        public object Default
        {
            get { return (defaultVal == null) ? null : Convert.ChangeType(defaultVal, propInfo.PropertyType); }
        }

        /// <summary>
        /// Retrieves the current value of the parameter for the specified instance.
        /// </summary>
        /// <param name="instance">The object to examine the parameter of.</param>
        /// <returns>The current value of the parameter.</returns>
        public object GetValue(object instance)
        {
            return propInfo.GetValue(instance, null);
        }

        /// <summary>
        /// Checks to see if the value provided lies in the allowed range for the parameter
        /// </summary>
        /// <param name="value">The value to test</param>
        /// <returns>
        /// <c>null</c> if no minimum or maximum is given, or if the parameter type is does not implement IComparable[itself], otherwise<para/>
        /// <c>true</c> if the value does lie between between <see cref="Minimum"/> and <see cref="Maximum"/>, otherwise false<para/>
        /// </returns>
        public bool? IsInRange(object value)
        {
            object convVal = Convert.ChangeType(value, propInfo.PropertyType);
            if (propertyTypeComparer != null && minVal != null && maxVal != null)
            {
                int minCheck = (int)propertyTypeComparer.Invoke(convVal, new object[] { Convert.ChangeType(minVal, propInfo.PropertyType) });
                int maxCheck = (int)propertyTypeComparer.Invoke(convVal, new object[] { Convert.ChangeType(maxVal, propInfo.PropertyType) });

                return (minCheck >= 0) && (maxCheck <= 0);
            }
            return null;
        }

        /// <summary>
        /// Attempts to set the value of the parameter for the instance specified.
        /// </summary>
        /// <param name="instance">The object to set the parameter of.</param>
        /// <param name="value">The value to attempt to set the parameter to.</param>
        /// <remarks>
        /// This method does not range-check the parameter, even if a range was specified.
        /// </remarks>
        public void SetValueUnchecked(object instance, object value)
        {
            propInfo.SetValue(instance, Convert.ChangeType(value, propInfo.PropertyType), null);
        }

        /// <summary>
        /// Attempts to set the value of the parameter for the instance specified.
        /// </summary>
        /// <param name="instance">The object to set the parameter of.</param>
        /// <param name="value">The value to attempt to set the parameter to.</param>
        /// <remarks>
        /// This method does range-check the parameter (for lying between <see cref="Minimum"/> and <see cref="Maximum"/>),
        /// if a range was specified.
        /// <para/>Throws System.ArgumentOutOfRangeException if the value does not lie between 
        /// <see cref="Minimum"/> and <see cref="Maximum"/></remarks>
        public void SetValue(object instance, object value)
        {
            if(!IsInRange(value).GetValueOrDefault(true))
                throw new ArgumentOutOfRangeException("value", String.Format(
                    "The parameter {0} cannot be set to {1}, it is outside the allowed range [{2} , {3}]",
                    name, value, minVal, maxVal));

            propInfo.SetValue(instance, Convert.ChangeType(value, propInfo.PropertyType), null);
        }

        /// <summary>
        /// Attempts to set the value of the parameter for the instance specified.
        /// </summary>
        /// <param name="instance">The object to set the parameter of.</param>
        /// <param name="value">The value to attempt to set the parameter to.</param>
        /// <remarks>
        /// This method does range-check <paramref name="value"/> (for lying between <see cref="Minimum"/> and <see cref="Maximum"/>),
        /// if a range was specified. If <paramref name="value"/> fails the range check, it is limited to <see cref="Minimum"/> or 
        /// <see cref="Maximum"/> as appropriate.</remarks>
        public void SetValueLimited(object instance, object value)
        {
            object convVal = Convert.ChangeType(value, propInfo.PropertyType);

            // Cam only make the comparison if the comparer method was found, and the maximum and minimum are appropriate
            if (propertyTypeComparer != null && minVal != null && maxVal != null)
            {
                int minCheck = (int)propertyTypeComparer.Invoke(convVal, new object[] { Convert.ChangeType(minVal, propInfo.PropertyType) });
                int maxCheck = (int)propertyTypeComparer.Invoke(convVal, new object[] { Convert.ChangeType(maxVal, propInfo.PropertyType) });

                if (minCheck < 0)
                    propInfo.SetValue(instance, Convert.ChangeType(minVal, propInfo.PropertyType), null);
                else if(maxCheck > 0)
                    propInfo.SetValue(instance, Convert.ChangeType(maxVal, propInfo.PropertyType), null);
                else
                    propInfo.SetValue(instance, convVal, null);
            }
            else propInfo.SetValue(instance, convVal, null);
        }

        /// <summary>
        /// Retrieves the current value of the parameter for the instance specified.
        /// </summary>
        /// <typeparam name="T">The type of object to retrie (should be assignable from the type given 
        /// by <see cref="ParamType"/>).</typeparam>
        /// <param name="instance">The object to examine the parameter of.</param>
        /// <returns>The current value of the parameter.</returns>
        public T GetValue<T>(object instance)
        {
            return (T)Convert.ChangeType(propInfo.GetValue(instance, null), typeof(T));
        }

        /// <summary>
        /// Attempts to set the (typed) value of the parameter for the instance specified.
        /// </summary>
        /// <typeparam name="T">Type of the value to set.</typeparam>
        /// <param name="instance">The object to set the parameter of.</param>
        /// <param name="value">The value to attempt to set the parameter to.</param>
        /// <remarks>
        /// This method does range-check the parameter (for lying between <see cref="Minimum"/> and <see cref="Maximum"/>),
        /// if a range was specified.
        /// <para/>Throws System.ArgumentOutOfRangeException if the value does not lie between 
        /// <see cref="Minimum"/> and <see cref="Maximum"/>
        /// <para/>If the type of parameter is not statically known, use <see cref="SetValue"/> instead.</remarks>
        public void SetValue<T>(object instance, T value) where T : IComparable<T>
        {
            if ((Minimum != null && value.CompareTo((T)Minimum) < 0) || (Maximum != null && value.CompareTo((T)Maximum) > 0))
                throw new ArgumentOutOfRangeException("value", String.Format(
                    "The parameter {0} cannot be set to {1}, it is outside the allowed range [{2} , {3}]", 
                    name, value, minVal, maxVal));

            propInfo.SetValue(instance, value, null);
        }
    }
}
