﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Diagnostics;

namespace VideoLib.Parameters
{
    /// <summary>
    /// Contains a number of static methods to help reflecting on properties marked with [<see cref="ParameterAttribute"/>].
    /// </summary>
    public static class ParameterReflector
    {
        // Returns the friendly name, if given, otherwise the property name with spaces inserted
        // before each upper-case letter

        /// <summary>
        /// Returns the friendly name, if given, otherwise the property name with spaces inserted before each upper-case letter
        /// </summary>
        /// <param name="fName">Friendly name, if any.</param>
        /// <param name="propName">Property name.</param>
        private static string GetName(string fName, string propName)
        {
            // Return the friendly name if there is one
            if (fName != null)
                return fName;

            // Do some helpful pascal-case reformatting (split the words on capital letter)
            StringBuilder sb = new StringBuilder();
            for (int index = 0; index < propName.Length; ++index)
            {
                // add a space before every captial letter except in the first word 
                if (Char.IsUpper(propName[index]) && index > 0)
                    sb.Append(' ');

                sb.Append(propName[index]);
            }
            return sb.ToString();
        }

        /// <summary>
        /// Enumerates all of the marked-up parameters in a single type.
        /// </summary>
        /// <param name="t">The type to search in.</param>
        /// <returns>
        /// A list of parameter information structures (which can be used to get/set the parameter value).
        /// </returns>
        public static List<ParameterInfo> EnumerateParameters(Type t)
        {
            return EnumerateParameters(t, true);
        }

        /// <summary>
        /// Enumerates all of the marked-up parameters in a single type.
        /// </summary>
        /// <param name="t">The type to search in.</param>
        /// <param name="traceHierarchy">If <c>true</c>, the inheritance hierarchy is searched for parameter properties</param>
        /// <returns>
        /// A list of parameter information structures (which can be used to get/set the parameter value).
        /// </returns>
        public static List<ParameterInfo> EnumerateParameters(Type t, bool traceHierarchy)
        {
            List<ParameterInfo> result = new List<ParameterInfo>();

            BindingFlags bindF = BindingFlags.Instance | BindingFlags.Public;
            if (!traceHierarchy)
                bindF |= BindingFlags.DeclaredOnly;

            // Find all of the potentially marked properties
            PropertyInfo[] pinfo = t.GetProperties(bindF);

            // Iterate through all parameters with getters and setters to find those marked
            // with the relevant custom attribute
            foreach (PropertyInfo pi in pinfo)
                if (pi.CanRead && pi.CanWrite)
                {
                    object[] attr = pi.GetCustomAttributes(typeof(ParameterAttribute), true);

                    if (attr != null && attr.Length > 0)
                    {
                        ParameterAttribute param = (ParameterAttribute)attr[0];

                        result.Add(new ParameterInfo(pi, GetName(param.FriendlyName, pi.Name), param.Description, param.GroupName,
                            param.Minimum, param.Maximum, param.Default));
                    }
                }

            return result;
        }

        /// <summary>
        /// Attempts to find a parameter in '<paramref name="t"/>' attached to the property "<paramref name="name"/>".
        /// If the parameter is not found, it generates a <see cref="ParameterNotFoundException"/>.
        /// </summary>
        /// <param name="t">The type to search in.</param>
        /// <param name="name">The CLR name of the property to find.</param>
        /// <returns>A parameter information structure, containing the details of the parameter.</returns>
        public static ParameterInfo FindParameter(Type t, string name)
        {
            PropertyInfo info = t.GetProperty(name);

            // This fails if the property is not a member of Type 't'
            if (info == null)
                throw new ParameterNotFoundException(t, name);

            object[] attr = info.GetCustomAttributes(typeof(ParameterAttribute), true);

            // This fails if the property has not been marked with the custom attribute
            if (attr == null || attr.Length <= 0)
                throw new ParameterNotFoundException(t, name);

            ParameterAttribute param = (ParameterAttribute)attr[0];

            return new ParameterInfo(info, GetName(param.FriendlyName, info.Name), param.Description, param.GroupName,
                param.Minimum, param.Maximum, param.Default);
        }

        /// <summary>
        /// Attempts to find a parameter in '<paramref name="t"/>' attached to the property "<paramref name="name"/>".
        /// If the parameter is not found, it does not set <paramref name="paramInfo"/>, and returns <c>false</c>.
        /// </summary>
        /// <param name="t">The type to search in.</param>
        /// <param name="name">The CLR name of the property to find.</param>
        /// <param name="paramInfo">A parameter information structure, set to contain the details of the parameter, if found.</param>
        /// <returns><c>true</c> if the parameter was found, otherwise <c>false</c>.</returns>
        public static bool TryFindParameter(Type t, string name, ref ParameterInfo paramInfo)
        {
            PropertyInfo info = t.GetProperty(name);

            // This fails if the property is not a member of Type 't'
            if (info == null)
                return false;

            object[] attr = info.GetCustomAttributes(typeof(ParameterAttribute), true);

            // This fails if the property has not been marked with the custom attribute
            if (attr == null || attr.Length <= 0)
                return false;

            ParameterAttribute param = (ParameterAttribute)attr[0];

            paramInfo = new ParameterInfo(info, GetName(param.FriendlyName, info.Name), param.Description, param.GroupName,
                param.Minimum, param.Maximum, param.Default);

            return true;
        }

        /// <summary>
        /// Finds all of the parameters for the specified object(s), and sets them to their default values, if given. 
        /// The objects passed must all be of the same type.
        /// </summary>
        /// <param name="objects">The objects to set the value of.</param>
        public static void ApplyDefaults(params object[] objects)
        {
            if (objects != null && objects.Length > 0)
            {
                Type commonType = objects[0].GetType();
                CheckTypes(commonType, objects);

                // Iterate through all of the properties, setting any 'parameters' to their default values
                PropertyInfo[] info = commonType.GetProperties();
                foreach (PropertyInfo pi in info)
                {
                    object[] attr = pi.GetCustomAttributes(typeof(ParameterAttribute), true);

                    // If the property is marked up as a parameter, set the value to the default
                    if (attr != null && attr.Length > 0)
                    {
                        ParameterAttribute param = (ParameterAttribute)attr[0];
                        foreach(object obj in objects)
                            pi.SetValue(obj, Convert.ChangeType(param.Default, pi.PropertyType), null);
                    }
                }
            }
        }

        [Conditional("DEBUG")]
        private static void CheckTypes(Type type, object[] objects)
        {
            // Ensure all other objects are of the same type
            for (int i = 0; i < objects.Length; ++i)
                if (!objects[i].GetType().Equals(type))
                    throw new InvalidOperationException("The objects passed are not all of the same type");
        }
    }
}
