﻿// $Id: WorkQueue.cs 487 2010-11-08 21:03:19Z cr333 $
// Partially based on a Part II project by Rubin Xu (2010).
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Threading;
using VideoLib.Frames;

namespace VideoLib.Scheduling
{
	/// <summary>Abstract base class for all work queues.</summary>
	public abstract class WorkQueue
	{
		/// <summary>The scheduler this queue is attached to.</summary>
		public Scheduler Scheduler { get; set; }

		/// <summary>The topology the scheduler is operating on.</summary>
		public Topology Topology { get; set; }

		/// <summary>The ready queue for tasks.</summary>
		protected readonly Queue<WorkItem> Ready = new Queue<WorkItem>();

		/// <summary>The worker threads associated with this work queue.</summary>
		protected List<Thread> workers = new List<Thread>();

//=================================================================================================
		#region Work queue interface

		/// <summary>Starts all worker threads associated with this work queue.</summary>
		/// <returns>Collection of all started worker threads.</returns>
		public abstract IEnumerable<Thread> StartWorkers();

		/// <summary>Convenience function to start a new thread.</summary>
		/// <param name="workerID">The worker ID passed to <see cref="WorkQueue.RunWorker"/>.</param>
		/// <param name="name">The thread name.</param>
		/// <returns>The thread handle of the new thread.</returns>
		protected Thread StartWorker(int workerID, string name)
		{
			// start a new scheduler thread
			Thread worker = new Thread(RunWorker) { Name = name };
			worker.Start(workerID);
			workers.Add(worker);
			return worker;
		}

		/// <summary>Main function of a worker thread. This is where the scheduling happens.</summary>
		/// <param name="data">The thread's worker ID.</param>
		protected virtual void RunWorker(object data)
		{
			// for profiling idle time
			long idletime = 0;

			// run an infinite loop, looking for work to do
			while (Scheduler.Status == SchedulerStatus.Running)
			{
				WorkItem workitem = null;

				lock (Ready)
				{
					// check if there is work to do (first chance)
					while (Ready.Count == 0 && Scheduler.Status == SchedulerStatus.Running)
					{
						if (Scheduler.Profiling) idletime -= Stopwatch.GetTimestamp();
						Monitor.Wait(Ready, 100);
						if (Scheduler.Profiling) idletime += Stopwatch.GetTimestamp();
					}

					// check if there is work to do (second chance)
					if (Ready.Count == 0 || Scheduler.Status != SchedulerStatus.Running)
						continue;

					// get work item to run
					workitem = Ready.Dequeue();
				}

				// run the work item
				workitem.Run();
			}

			if (Scheduler.Profiling) Debug.WriteLine(string.Format(Thread.CurrentThread.Name + ": Stopped. Idle time: {0:F3} s", (double)idletime / Stopwatch.Frequency));
			else                     Debug.WriteLine(Thread.CurrentThread.Name + ": Stopped.");
		}

		/// <summary>Handles initialisation of nodes.</summary>
		/// <param name="node">The node to be initialised.</param>
		/// <returns>True, if and only if the node has been initialised.</returns>
		public virtual bool HandleNodeInitEvent(INode node)
		{
			// default: event not handled
			return false;
		}

		/// <summary>Handles new frame events.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public virtual bool HandleNewFrameEvent(object sender, NewFrameEventArgs e)
		{
			// default: event not handled
			return false;
		}

		/// <summary>Handles nodes that become ready to run.</summary>
		/// <param name="node">The ready node.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public virtual bool HandleNodeReadyEvent(INode node)
		{
			// default: event not handled
			return false;
		}

		/// <summary>Handles changes of pin status.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="e">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public virtual bool HandlePinStatusChangedEvent(object sender, PinStatusChangedEventArgs e)
		{
			// default: event not handled
			return false;
		}

		#endregion // Work queue interface
//=================================================================================================
		#region Static reference implementations

		/// <summary>Handles initialisation of nodes.</summary>
		/// <param name="node">The node to be initialised.</param>
		/// <param name="readyQueue">The work queue.</param>
		/// <returns>True, if and only if the node has been initialised.</returns>
		public static bool HandleNodeInitEvent(INode node, Queue<WorkItem> readyQueue)
		{
			lock (readyQueue)
			{
				// add work item to initialise node on the worker thread
				readyQueue.Enqueue(new ActionWorkItem(delegate()
				{
					node.Init();
					node.UpdateStatus();
				}));
				Monitor.Pulse(readyQueue);
			}

			return true;
		}

		/// <summary>Handles new frame events using transport work items.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="eventArgs">The <see cref="VideoLib.Frames.NewFrameEventArgs"/> instance containing the event data.</param>
		/// <param name="topology">The topology of nodes.</param>
		/// <param name="readyQueue">The work queue.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public static bool HandleNewFrameEvent(object sender, NewFrameEventArgs eventArgs, Topology topology, Queue<WorkItem> readyQueue)
		{
			OutputPin outputPin = (OutputPin)sender;

			if (!topology.OutgoingEdges.Keys.Contains(outputPin))
				return false;

			// need to increment reference count appropriately, if applicable
			IRefCountedFrame frame = eventArgs.Frame as IRefCountedFrame;

			foreach (var inputPin in topology.OutgoingEdges[outputPin])
			{
				lock (readyQueue)
				{
					// increment reference count for each outgoing edge
					if (frame != null) frame.RefInc();

					readyQueue.Enqueue(new TransportWorkItem(outputPin, inputPin, eventArgs));

					// wake up a worker thread to process this new item.
					Monitor.Pulse(readyQueue);
				}
			}

			return true; // event handled
		}

		/// <summary>Handles nodes that become ready to run.</summary>
		/// <param name="node">The ready node.</param>
		/// <param name="profiler">The profiler to use (or null).</param>
		/// <param name="readyQueue">The work queue.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public static bool HandleNodeReadyEvent(INode node, IProfiler profiler, Queue<WorkItem> readyQueue)
		{
			lock (readyQueue)
			{
				readyQueue.Enqueue(new ProcessingWorkItem(node) { Profiler = profiler });
				Monitor.Pulse(readyQueue);
			}

			return true; // event handled
		}

		/// <summary>Handles changes of pin status.</summary>
		/// <param name="sender">The sender.</param>
		/// <param name="eventArgs">The <see cref="VideoLib.PinStatusChangedEventArgs"/> instance containing the event data.</param>
		/// <param name="topology">The topology of nodes.</param>
		/// <param name="readyQueue">The work queue.</param>
		/// <returns>True, if and only if the event was handled.</returns>
		public static bool HandlePinStatusChangedEvent(object sender, PinStatusChangedEventArgs eventArgs, Topology topology, Queue<WorkItem> readyQueue)
		{
			if (sender is InputPin) // then update upstream output pins
			{
				foreach (var outPin in topology.IncomingEdges[(InputPin)sender])
				{
					lock (readyQueue)
					{
						readyQueue.Enqueue(new ActionWorkItem(delegate() { outPin.DownStreamStatusChangedListener(sender, eventArgs); }));
						Monitor.Pulse(readyQueue);
					}
				}
				return true; // event handled
			}

			if (sender is OutputPin) // then update node
			{
				OutputPin outputPin = (OutputPin)sender;

				lock (readyQueue)
				{
					readyQueue.Enqueue(new ActionWorkItem(delegate() { outputPin.Node.UpdateStatus(); }));
					Monitor.Pulse(readyQueue);
				}
				return true; // event handled
			}

			return false; // event not handled
		}

		#endregion // Reference implementations
//=================================================================================================
	}
}